#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import errno
import inspect

import shutil
import tarfile

import re

from . import jsonfiles, string
from .errors import *
from .simulation import Simulation
from . import checkers

class Manager():
	'''
	Manage a simulations folder: add, delete, extract or update simulations, based on their settings.

	Parameters
	----------
	folder : Folder
		The folder to manage.
	'''

	def __init__(self, folder):
		self._folder = folder

		self._simulations_list_file = os.path.join(self._folder.folder, '.simulations.list')
		self._simulations_list_dict = None

		self._checkers_regex_compiled = None
		self._checkers_list = None

	@property
	def _simulations_list(self):
		'''
		Return the simulations list.

		Returns
		-------
		simulations_list : dict
			A name: settings dictionary.
		'''

		if self._simulations_list_dict is None:
			try:
				self._simulations_list_dict = jsonfiles.read(self._simulations_list_file)

			except FileNotFoundError:
				self._simulations_list_dict = {}

		return self._simulations_list_dict

	def saveSimulationsList(self):
		'''
		Save the list of simulations.
		'''

		jsonfiles.write(self._simulations_list, self._simulations_list_file)

	@property
	def _checkers_regex(self):
		'''
		Regex to determine whether a function's name corresponds to a checker.

		Returns
		-------
		regex : re.Pattern
			The checkers regex.
		'''

		if self._checkers_regex_compiled is None:
			self._checkers_regex_compiled = re.compile(r'^(file|folder|global)_([A-Za-z0-9_]+)$')

		return self._checkers_regex_compiled

	@property
	def _checkers(self):
		'''
		Get the list of available checkers.

		Returns
		-------
		checkers : dict
			Checkers, sorted by category.
		'''

		if self._checkers_list is None:
			self._checkers_list = {cat: {} for cat in ['file', 'folder', 'global']}
			self.loadCheckersFromModule(checkers)

		return self._checkers_list

	def loadCheckersFromModule(self, module):
		'''
		Load all checkers in a given module.

		Parameters
		----------
		module : Module
			Module (already loaded) where are defined the checkers.
		'''

		for function in inspect.getmembers(module, inspect.isfunction):
			checker_match = self._checkers_regex.match(function[0])

			if checker_match:
				self.setChecker(checker_match.group(1), checker_match.group(2), function[1])

	def setChecker(self, category, checker_name, checker):
		'''
		Set (add or replace) a checker.

		Parameters
		----------
		category : str
			Category of the checker (`file`, `folder` or `global`).

		checker_name : str
			Name of checker.

		checker : function
			Checker to register (callback function).
		'''

		if not(category in self._checkers):
			raise CheckersCategoryNotFoundError(category)

		self._checkers[category][checker_name] = checker

	def removeChecker(self, category, checker_name):
		'''
		Remove a checker.

		Parameters
		----------
		category : str
			Category of the checker (`file`, `folder` or `global`).

		checker_name : str
			Name of checker.
		'''

		if not(category in self._checkers):
			raise CheckersCategoryNotFoundError(category)

		if not(checker_name in self._checkers[category]):
			raise CheckerNotFoundError(checker_name, category)

		del self._checkers[category][checker_name]

	def checkIntegrity(self, simulation):
		'''
		Check the integrity of a simulation.

		Parameters
		----------
		simulation : Simulation
			The simulation to check.

		Returns
		-------
		success : bool
			`True` if the integrity check is successful, `False` otherwise.
		'''

		tree = {}

		for output_entry in ['files', 'folders']:
			tree[output_entry] = []
			checkers_cat = output_entry[:-1]

			if output_entry in self._folder.settings['output']:
				for output in self._folder.settings['output'][output_entry]:
					parsed_name = str(simulation.parseString(output['name']))
					tree[output_entry].append(parsed_name)

					if 'checks' in output:
						for checker_name in output['checks']:
							if not(checker_name in self._checkers[checkers_cat]):
								raise CheckerNotFoundError(checker_name, checkers_cat)

							if not(self._checkers[checkers_cat][checker_name](simulation, parsed_name)):
								return False

		if 'checks' in self._folder.settings['output']:
			for checker_name in self._folder.settings['output']['checks']:
				if not(checker_name in self._checkers['global']):
					raise CheckerNotFoundError(checker_name, 'global')

				if not(self._checkers['global'][checker_name](simulation, tree)):
					return False

		return True

	def compress(self, folder, simulation_name):
		'''
		Create an archive to store a simulation.

		Parameters
		----------
		folder : str
			Folder to compress.

		simulation_name : str
			Name to use for the archive.
		'''

		with tarfile.open(os.path.join(self._folder.folder, f'{simulation_name}.tar.bz2'), 'w:bz2') as tar:
			tar.add(folder, arcname = simulation_name)

		shutil.rmtree(folder)

	def uncompress(self, simulation_name, folder):
		'''
		Extract a simulation from an archive.

		Parameters
		----------
		simulation_name : str
			Name of the archive to extract.

		folder : str
			Folder into which the files must go.
		'''

		with tarfile.open(os.path.join(self._folder.folder, f'{simulation_name}.tar.bz2'), 'r:bz2') as tar:
			tar.extractall(path = self._folder.folder)

		shutil.move(os.path.join(self._folder.folder, simulation_name), folder)

	def add(self, simulation, save_list = True):
		'''
		Add a simulation to the list.

		Parameters
		----------
		simulation : Simulation|dict
			The simulation to add.

		save_list : boolean
			`True` to save the simulations list, `False` otherwise.

		Raises
		------
		SimulationFolderNotFoundError
			The folder indicated in the simulation does not exist.

		SimulationIntegrityCheckFailedError
			At least one integrity check failed.
		'''

		simulation = Simulation.ensureType(simulation, self._folder)

		if not(os.path.isdir(simulation['folder'])):
			raise SimulationFolderNotFoundError(simulation['folder'])

		settings_str = string.fromObject(simulation.settings)
		simulation_name = string.hash(settings_str)

		if not(self.checkIntegrity(simulation)):
			raise SimulationIntegrityCheckFailedError(simulation['folder'])

		self.compress(simulation['folder'], simulation_name)

		self._simulations_list[simulation_name] = settings_str

		if save_list:
			self.saveSimulationsList()

	def delete(self, simulation, save_list = True):
		'''
		Delete a simulation.

		Parameters
		----------
		simulation : Simulation|dict
			The simulation to delete.

		save_list : boolean
			`True` to save the simulations list, `False` otherwise.

		Raises
		------
		SimulationNotFoundError
			The simulation does not exist in the list.
		'''

		simulation = Simulation.ensureType(simulation, self._folder)
		simulation_name = string.hash(string.fromObject(simulation.settings))

		if not(simulation_name in self._simulations_list):
			raise SimulationNotFoundError(simulation_name)

		os.unlink(os.path.join(self._folder.folder, f'{simulation_name}.tar.bz2'))
		del self._simulations_list[simulation_name]

		if save_list:
			self.saveSimulationsList()

	def extract(self, simulation):
		'''
		Extract a simulation.

		Parameters
		----------
		simulation : Simulation|dict
			The simulation to extract.

		Raises
		------
		SimulationNotFoundError
			The simulation does not exist in the list.

		SimulationFolderAlreadyExistError
			The destination of extraction already exists.
		'''

		simulation = Simulation.ensureType(simulation, self._folder)
		simulation_name = string.hash(string.fromObject(simulation.settings))

		if not(simulation_name in self._simulations_list):
			raise SimulationNotFoundError(simulation_name)

		if os.path.exists(simulation['folder']):
			raise SimulationFolderAlreadyExistError(simulation['folder'])

		destination_path = os.path.dirname(os.path.normpath(simulation['folder']))
		if destination_path and not(os.path.isdir(destination_path)):
			os.makedirs(destination_path)

		self.uncompress(simulation_name, simulation['folder'])

	def batchAction(self, simulations, action, args = {}, *, save_list = True, errors_store = (), errors_pass = (Error), callback = None):
		'''
		Apply a callback function to each simulation of a given list.

		Parameters
		----------
		simulations : list
			List of simulations.

		action : function
			Function to call. The simulation will be passed as the first parameter.

		args : dict
			Additional named arguments to pass to the callback.

		save_list : boolean
			`True` to save the simulations list once the loop is over, `False` to not save it.

		errors_store : tuple
			List of exceptions that, when raised, lead to the storage of the involved simulation.

		errors_pass : tuple
			List of exceptions that, when raised, do nothing.

		callback : function
			Function to call at the end of each action.

		Returns
		-------
		errors : list
			List of simulations which raised an error.
		'''

		errors = []

		for simulation in simulations:
			try:
				action(simulation, **args)

			except errors_store:
				errors.append(simulation)

			except errors_pass:
				pass

			if not(callback is None):
				callback()

		if save_list:
			self.saveSimulationsList()

		return errors

	def batchAdd(self, simulations, *, callback = None):
		'''
		Add multiple simulations to the list.

		Parameters
		----------
		simulations : list
			List of simulations.

		callback : function
			Function to call at the end of each addition.

		Returns
		-------
		errors : list
			List of simulations that were not added because they raised an error.
		'''

		return self.batchAction(simulations, self.add, {'save_list': False}, save_list = True, errors_store = (SimulationFolderNotFoundError, SimulationIntegrityCheckFailedError), callback = callback)

	def batchDelete(self, simulations, *, callback = None):
		'''
		Delete multiple simulations.

		Parameters
		----------
		simulations : list
			List of simulations.

		callback : function
			Function to call at the end of each deletion.

		Returns
		-------
		errors : list
			List of simulations that were not deleted because they raised an error.
		'''

		return self.batchAction(simulations, self.delete, {'save_list': False}, save_list = True, errors_store = (SimulationNotFoundError), callback = callback)

	def batchExtract(self, simulations, *, ignore_existing = True, callback = None):
		'''
		Extract multiple simulations.

		Parameters
		----------
		simulations : list
			List of simulations.

		ignore_existing : boolean
			Ignore simulations for which the destination folder already exists.

		callback : function
			Function to call at the end of each extraction.

		Returns
		-------
		errors : list
			List of simulations that were not extracted because they raised an error.
		'''

		if ignore_existing:
			errors_store = (SimulationNotFoundError,)
			errors_pass = (SimulationFolderAlreadyExistError,)

		else:
			errors_store = (SimulationNotFoundError, SimulationFolderAlreadyExistError)
			errors_pass = ()

		return self.batchAction(simulations, self.extract, save_list = False, errors_store = errors_store, errors_pass = errors_pass, callback = callback)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import functools
import re
import inspect

from .errors import *
from . import fixers

class Simulation():
	'''
	Represent a simulation, itself identified by its settings.

	Parameters
	----------
	folder : Folder
		Instance of `Folder` for the current simulation's folder.

	settings : dict
		Dictionary listing the user settings.
	'''

	def __init__(self, folder, settings):
		self._folder = folder
		self._user_settings = settings

		self._raw_globalsettings = None
		self._raw_settings = None

		self._fixers_regex_compiled = None
		self._fixers_list = None

		self._setting_tag_regex_compiled = None
		self._parser_recursion_stack = []

	@classmethod
	def ensureType(cls, simulation, folder):
		'''
		Ensure a variable is a Simulation object.

		Parameters
		----------
		simulation : dict|Simulation
			The variable to check.

		folder : Folder
			The `Folder` instance to use in case we need to create a new object.

		Returns
		-------
		simulation : Simulation
			The simulation as a Simulation instance.
		'''

		if type(simulation) is cls:
			return simulation

		return cls(folder, simulation)

	def __getitem__(self, key):
		'''
		Access to a global setting.

		Parameters
		----------
		key : str
			The key of the setting to get.

		Raises
		------
		KeyError
			The key does not exist.

		Returns
		-------
		value : mixed
			The corresponding value.
		'''

		try:
			return self.reduced_globalsettings[key]

		except KeyError:
			raise KeyError('The key does not exist in the global settings')

	def __setitem__(self, key, value):
		'''
		Change a global setting.

		Parameters
		----------
		key : str
			The key of the setting to change.

		value : mixed
			The new value of the setting.

		Raises
		------
		KeyError
			The key does not exist.
		'''

		try:
			setting = [setting for setting in self._globalsettings if setting['name'] == key][0]

		except IndexError:
			raise KeyError('The key does not exist in the global settings')

		else:
			setting['value'] = value

	@property
	def _globalsettings(self):
		'''
		Return (and generate if needed) the complete list of global settings.

		Returns
		-------
		raw_globalsettings : list
			The global settings.
		'''

		if not(self._raw_globalsettings):
			self.generateGlobalSettings()

		return self._raw_globalsettings

	@property
	def _settings(self):
		'''
		Return (and generate if needed) the complete list of settings.

		Returns
		-------
		raw_settings : list
			The settings.
		'''

		if not(self._raw_settings):
			self.generateSettings()

		return self._raw_settings

	@property
	def settings(self):
		'''
		Return the complete list of sets of settings to use, as dictionaries.
		The settings with `exclude` to `True` are ignored.

		Returns
		-------
		settings : list
			List of sets of settings.
		'''

		return [
			{s['name']: s['value'] for s in settings_set if not(s['exclude'])}
			for settings_set in self._settings
		]

	@property
	def settings_as_strings(self):
		'''
		Return the complete list of sets of settings to use, as strings.

		Returns
		-------
		settings : list
			Settings, generated according to their pattern.
		'''

		return [
			[s['pattern'].format(name = s['name'], value = s['value']) for s in settings_set]
			for settings_set in self._settings
		]

	@property
	def reduced_globalsettings(self):
		'''
		Return the list of global settings, as a name: value dictionary.

		Returns
		-------
		settings : dict
			The global settings.
		'''

		return {setting['name']: setting['value'] for setting in self._globalsettings}

	@property
	def reduced_settings(self):
		'''
		Return the list of settings, as a name: value dictionary.
		Ignore multiple occurrences of the same setting.

		Returns
		-------
		settings : dict
			The settings.
		'''

		return functools.reduce(lambda a, b: {**a, **b}, self.settings)

	@property
	def command_line(self):
		'''
		Return the command line to use to generate this simulation.

		Returns
		-------
		command_line : str
			The command line to execute.
		'''

		return ' '.join([self._folder.settings['exec']] + sum(self.settings_as_strings, []))

	@property
	def _fixers_regex(self):
		'''
		Regex to detect whether a function's name corresponds to a value fixer.

		Returns
		-------
		regex : re.Pattern
			The fixers regex.
		'''

		if self._fixers_regex_compiled is None:
			self._fixers_regex_compiled = re.compile(r'^fixer_([A-Za-z0-9_]+)$')

		return self._fixers_regex_compiled

	@property
	def _setting_tag_regex(self):
		'''
		Regex to detect whether there is a setting or global setting tag in a string.

		Returns
		-------
		regex : re.Pattern
			The setting tag regex.
		'''

		if self._setting_tag_regex_compiled is None:
			self._setting_tag_regex_compiled = re.compile(r'\{(?P<category>(?:global)?setting):(?P<name>[^}]+)\}')

		return self._setting_tag_regex_compiled

	@property
	def _fixers(self):
		'''
		Get the list of available values fixers.

		Returns
		-------
		fixers : dict
			The values fixers.
		'''

		if self._fixers_list is None:
			self._fixers_list = {}
			self.loadFixersFromModule(fixers)

		return self._fixers_list

	def loadFixersFromModule(self, module):
		'''
		Load all values fixers in a given module.

		Parameters
		----------
		module : Module
			Module (already loaded) where are defined the fixers.
		'''

		for function in inspect.getmembers(module, inspect.isfunction):
			fixer_match = self._fixers_regex.match(function[0])

			if fixer_match:
				self.setFixer(fixer_match.group(1), function[1])

	def setFixer(self, fixer_name, fixer):
		'''
		Set (add or replace) a value fixer.

		Parameters
		----------
		fixer_name : str
			Name of the fixer.

		fixer : function
			Fixer to register.
		'''

		self._fixers[fixer_name] = fixer

	def removeFixer(self, fixer_name):
		'''
		Remove a value fixer.

		Parameters
		----------
		fixer_name : str
			Name of the fixer to remove.
		'''

		if not(fixer_name in self._fixers):
			raise FixerNotFoundError(fixer_name)

		del self._fixers[fixer_name]

	def generateGlobalSettings(self):
		'''
		Generate the full list of global settings.
		'''

		self._raw_globalsettings = []

		for setting in self._folder.settings['globalsettings']:
			self._raw_globalsettings.append({
				'name': setting['name'],
				'value': self.fixValue(self._user_settings[setting['name']]) if setting['name'] in self._user_settings else setting['default']
			})

		self.parseGlobalSettings()

	def generateSettings(self):
		'''
		Generate the full list of settings, taking into account the user settings and the default values in the folder.
		The "raw settings" are generated. Each setting is a dictionary:
		```
		{
			'name': 'name of the setting',
			'value': 'value of the setting (initialized with the default one)',
			'exclude': 'boolean to determine if we should exclude this setting for the comparison things',
			'pattern': 'the pattern to use when we generate the corresponding string'
		}
		```
		Each set of settings is a list of all settings in this set.
		'''

		self._raw_settings = []
		default_pattern = self._folder.settings['setting_pattern']

		for settings_set in self._folder.settings['settings']:
			default_settings = [
				{
					'name': s['name'],
					'value': s['default'],
					'exclude': 'exclude' in s and s['exclude'],
					'pattern': s['pattern'] if 'pattern' in s else default_pattern
				}
				for s in settings_set['settings']
			]

			values_sets = [s['settings'] for s in self._user_settings['settings'] if s['set'] == settings_set['set']]

			if values_sets:
				for values_set in values_sets:
					set_to_add = copy.deepcopy(default_settings)

					for setting in set_to_add:
						try:
							setting['value'] = self.fixValue(values_set[setting['name']])

						except KeyError:
							pass

					self._raw_settings.append(set_to_add)

			elif settings_set['required']:
				self._raw_settings.append(default_settings)

		self.parseSettings()

	def fixValue(self, value):
		'''
		Fix a value to prevent false duplicates (e.g. this prevent to consider `0.0` and `0` as different values).

		Parameters
		----------
		value : mixed
			The value to fix.

		Returns
		-------
		fixed : mixed
			The same value, fixed.
		'''

		if 'fixes' in self._folder.settings:
			for fixer in self._folder.settings['fixes']:
				if not(type(fixer) is list):
					fixer = [fixer]

				if not(fixer[0] in self._fixers):
					raise FixerNotFoundError(fixer[0])

				value = self._fixers[fixer[0]](value, *fixer[1:])

		return value

	def parseString(self, s):
		'''
		Parse a string to take into account possible settings.
		The tag `{setting:name}` is replaced by the value of the simulation's setting named `name`.
		The tag `{globalsetting:name}` is replaced by the value of the global setting named `name`.

		Tags are replaced recursively.

		Parameters
		----------
		s : str
			The string to parse.

		Returns
		-------
		parsed : mixed
			The parsed string, or a copy of the setting if the whole string is just one tag.
		'''

		if not(type(s) is str):
			return s

		settings = {
			'setting': self.reduced_settings,
			'globalsetting': self.reduced_globalsettings
		}

		fullmatch = self._setting_tag_regex.fullmatch(s)

		if fullmatch:
			try:
				return copy.deepcopy(settings[fullmatch.group('category')][fullmatch.group('name')])

			except KeyError:
				return s

		parsed = ''
		k0 = 0

		for match in self._setting_tag_regex.finditer(s):
			parsed += s[k0:match.start()]

			try:
				parsed += str(settings[match.group('category')][match.group('name')])

			except KeyError:
				parsed += match.group(0)

			k0 = match.end()

		parsed += s[k0:]

		self._parser_recursion_stack.append(s)

		if not(parsed in self._parser_recursion_stack):
			return self.parseString(parsed)

		self._parser_recursion_stack.clear()

		return parsed

	def parseGlobalSettings(self):
		'''
		Parse the global settings to take into account possible other settings' values.
		'''

		for setting in self._globalsettings:
			setting['value'] = self.parseString(setting['value'])

	def parseSettings(self):
		'''
		Parse the settings to take into account possible other settings' values.
		'''

		for settings_set in self._settings:
			for setting in settings_set:
				setting['value'] = self.parseString(setting['value'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import imaplib
import email

from . import jsonfiles
from .errors import *

class Watcher():
	'''
	Watch for events to detect the end of some simulations.

	Parameters
	----------
	remote_folder : RemoteFolder
		A RemoteFolder instance to use to search for notifications.

	states_path : str
		Path to the remote file where the states are stored.

	mail_config : dict
		Configuration of the mailbox.

	mail_notifications_config : dict
		Mails notifications configuration.

	Raises
	------
	WatcherNoConfigFoundError
		Nothing used to configure the watcher.

	WatcherTooManyConfigError
		More than one configuration used.
	'''

	def __init__(self, *, remote_folder = None, states_path = '', mail_config = None, mail_notifications_config = None):
		self._jobs_to_watch = set()
		self._jobs_states = {}

		self._remote_folder = remote_folder
		self._states_path = states_path

		self._mail_config = mail_config
		self._mail_notifications_config_dict = mail_notifications_config
		self._mail_notifications_config_compiled = False
		self._mailbox_instance = None

		if self._remote_folder is None and self._mail_config is None:
			raise WatcherNoConfigFoundError()

		if not(self._remote_folder is None) and not(self._mail_config is None):
			raise WatcherTooManyConfigError()

	@property
	def _mailbox(self):
		'''
		Get or create an object to access the mailbox.

		Raises
		------
		WatcherNoMailConfigError
			No configuration found for the mailbox.

		Returns
		-------
		mailbox : imaplib.IMAP4
			The mailbox object.
		'''

		if self._mailbox_instance is None:
			self._mailbox_instance = imaplib.IMAP4(self._mail_config['host'], self._mail_config['port'])
			self._mailbox_instance.login(self._mail_config['user'], self._mail_config['password'])
			self._mailbox_instance.select('INBOX')

		return self._mailbox_instance

	@property
	def _mail_notifications_config(self):
		'''
		Get or create the dictionary containing the mails notifications configuration.

		Raises
		------
		WatcherNoMailNotificationsConfigError
			No configuration found fot the notifications.

		Returns
		-------
		config : dict
			The mail notifications configuration.
		'''

		if not(self._mail_notifications_config_compiled):
			self._mail_notifications_config_dict['search'] = '"' + self._mail_notifications_config_dict['search'] + '"'
			self._mail_notifications_config_dict['infos_extraction']['subject_regex'] = re.compile(self._mail_notifications_config_dict['infos_extraction']['subject'])

			self._mail_notifications_config_compiled = True

		return self._mail_notifications_config_dict

	def close(self):
		'''
		Close the mailbox if necessary.
		'''

		if not(self._mailbox_instance is None):
			self._mailbox_instance.logout()

	def addJobsToWatch(self, jobs):
		'''
		Add some jobs to watch.

		Parameters
		----------
		jobs : list
			IDs of the jobs.
		'''

		self._jobs_to_watch |= set(jobs)

	def clearJobs(self):
		'''
		Clear the list of jobs to watch.
		'''

		self._jobs_to_watch.clear()
		self._jobs_states.clear()

	def getJobsByStates(self, states):
		'''
		Get the list of jobs which are in a given state.

		Parameters
		----------
		states : list
			List of states to use to filter the jobs.

		Returns
		-------
		jobs : list
			The list of jobs in these states.
		'''

		return [job for job, state in self._jobs_states.items() if state in states]

	def getNumberOfJobsByState(self, state):
		'''
		Return the number of jobs in a given state.

		Parameters
		----------
		state : str
			The state to use to filter the jobs.

		Returns
		-------
		n_jobs : int
			The number of jobs in this state.
		'''

		return len(self.getJobsByStates([state]))

	def getNumberOfJobsByStates(self, states):
		'''
		Return the number of jobs in given states.

		Parameters
		----------
		states : list
			The states to use to filter the jobs.

		Returns
		-------
		n_jobs : dict
			A dictionary with keys equal to the wanted states and values the number of jobs in these states.
		'''

		return {state: self.getNumberOfJobsByState(state) for state in states}

	def setJobsStatesPath(self, states_path):
		'''
		Set the path to the remote file where we can find the jobs states.

		Parameters
		----------
		states_path : str
			Path to the remote file containing the jobs states.

		Raises
		------
		WatcherNoRemoteFolderError
			There is no RemoteFolder set.
		'''

		if self._remote_folder is None:
			raise WatcherNoRemoteFolderError()

		self._states_path = states_path

	def setMailNotificationsConfig(self, config):
		'''
		Set the mails notifications configuration file.

		Parameters
		----------
		config : str
			Path to the file containing the configuration.

		Raises
		------
		WatcherNoMailConfigError
			There is no mailbox set.
		'''

		if self._mail_config is None:
			raise WatcherNoMailConfigError()

		self._mail_notifications_config_file = config

	def _getJobsStatesFromFile(self):
		'''
		Remotely read the current states of the jobs.

		Raises
		------
		WatcherNoStatesPathError
			There is no path to look for the states.

		Returns
		-------
		known_states : dict
			A dictionary listing the current jobs states.
		'''

		if not(self._states_path):
			raise WatcherNoStatesPathError()

		known_states = {}

		try:
			known_states = dict(self._remote_folder.getFileContents(self._states_path, callback = lambda l: tuple(map(lambda s: s.strip(), l.split(':')))))

		except FileNotFoundError:
			pass

		return known_states

	def _getJobsStatesFromMail(self):
		'''
		Search for mail notifications to update the jobs states.

		Returns
		-------
		known_states : dict
			A dictionary listing the current jobs states.
		'''

		known_states = {}

		typ, data_nums = self._mailbox.search(None, 'SUBJECT', self._mail_notifications_config['search'])

		for num in data_nums[0].split():
			typ, data = self._mailbox.fetch(num, '(RFC822)')
			msg = email.message_from_bytes(data[0][1])
			subject = email.header.decode_header(msg['Subject'])[0][0]

			matches = self._mail_notifications_config['infos_extraction']['subject_regex'].match(subject)
			job_id = matches.group('id')

			if job_id in self._jobs_to_watch:
				job_state = matches.group('state')

				try:
					job_state = self._mail_notifications_config['convert'][job_state]

				except KeyError:
					pass

				known_states[job_id] = job_state

				if 'delete' in self._mail_notifications_config and self._mail_notifications_config['delete']:
					self._mailbox.store(num, '+FLAGS', '\\Deleted')

		self._mailbox.expunge()

		return known_states

	def updateJobsStates(self):
		'''
		Read the current states of the jobs.
		'''

		known_states = {}

		if not(self._remote_folder is None):
			known_states = self._getJobsStatesFromFile()

		else:
			known_states = self._getJobsStatesFromMail()

		for job in self._jobs_to_watch & set(known_states.keys()):
			state = known_states[job]

			if state in ['waiting', 'running', 'succeed', 'failed']:
				self._jobs_states[job] = state

		self._jobs_states.update({job: 'waiting' for job in self._jobs_to_watch - set(self._jobs_states.keys())})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class Error(Exception):
	'''
	Base class for exceptions occurring in the manager.
	'''

	pass

class SimulationFolderNotFoundError(Error):
	'''
	Exception raised when we try to add a simulation to the manager, but the indicated folder does not exist.

	Parameters
	----------
	folder : str
		The folder which has not been found.
	'''

	def __init__(self, folder):
		self.folder = folder

class CheckersCategoryNotFoundError(Error):
	'''
	Exception raised when we try to access a checkers category which does not exist.

	Parameters
	----------
	category : str
		The category which has not been found.
	'''

	def __init__(self, category):
		self.category = category

class CheckerNotFoundError(Error):
	'''
	Exception raised when we try to access a checker which does not exist.

	Parameters
	----------
	category : str
		The checker's category.

	checker_name : str
		The name of the checker which has not been found.
	'''

	def __init__(self, category, checker_name):
		self.category = category
		self.checker_name = checker_name

class VariableGeneratorNotFoundError(Error):
	'''
	Exception raised when we try to access a variable generator which does not exist.

	Parameters
	----------
	generator_name : str
		Name of the generator which has not been found.
	'''

	def __init__(self, generator_name):
		self.generator_name = generator_name

class FixerNotFoundError(Error):
	'''
	Exception raised when we try to access a value fixer which does not exist.

	Parameters
	----------
	fixer_name : str
		Name of the fixer which has not been found.
	'''

	def __init__(self, fixer_name):
		self.fixer_name = fixer_name

class SimulationIntegrityCheckFailedError(Error):
	'''
	Exception raised when a folder fails to pass an integrity check.

	Parameters
	----------
	folder : str
		The folder which has not been found.
	'''

	def __init__(self, folder):
		self.folder = folder

class SimulationNotFoundError(Error):
	'''
	Exception raised when we look for a non existing simulation.

	Parameters
	----------
	simulation : str
		The name of the simulation which has not been found.
	'''

	def __init__(self, simulation):
		self.simulation = simulation

class SimulationFolderAlreadyExistError(Error):
	'''
	Exception raised when the folder of a simulation already exists.

	Parameters
	----------
	folder : str
		The name of the folder which should not exist.
	'''

	def __init__(self, folder):
		self.folder = folder

class EmptyListError(Error):
	'''
	Exception raised when a specific list is empty.
	'''

	pass

class DestinationFolderExistsError(Error):
	'''
	Exception raised when a destination folder already exists.
	'''

	pass

class ScriptNotFoundError(Error):
	'''
	Exception raised when a given script does not exist.

	Parameters
	----------
	script_coords : dict
		Coordinates of the script which has not been found.
	'''

	def __init__(self, script_coords):
		self.script_coords = script_coords

class RemotePathNotFoundError(Error):
	'''
	Exception raised when we try to access a remote path/directory which does not exist.

	Parameters
	----------
	remote_path : str
		The path.
	'''

	def __init__(self, remote_path):
		self.remote_path = remote_path

class UILineNotFoundError(Error):
	'''
	Exception raised when we try to access to a UI line which does not exist.

	Parameters
	----------
	id : str
		The ID of the unknown line.
	'''

	def __init__(self, id):
		self.id = id

class UIProgressBarNotFoundError(Error):
	'''
	Exception raised when we try to access to a UI progress bar which does not exist.

	Parameters
	----------
	id : str
		The ID of the unknown progress bar.
	'''

	def __init__(self, id):
		self.id = id

class UINonMovableLine(Error):
	'''
	Exception raised when we try to move a line which can't be moved.

	Parameters
	----------
	pos : int
		Position of the line.
	'''

	def __init__(self, pos):
		self.pos = pos

class WatcherNoConfigFoundError(Error):
	'''
	Exception raised when nothing is used to configure the watcher.
	'''

	pass

class WatcherTooManyConfigError(Error):
	'''
	Exception raised when there is more than one configuration for the watcher.
	'''

	pass

class WatcherNoRemoteFolderError(Error):
	'''
	Exception raised when we try to use the watcher's remote folder while there is no one set.
	'''

	pass

class WatcherNoStatesPathError(Error):
	'''
	Exception raised when we try to access the file containing the jobs states and there is no one set.
	'''

	pass

class WatcherNoMailConfigError(Error):
	'''
	Exception raised when we try to connect to the mailbox and no configuration file is given.
	'''

	pass

class WatcherNoMailNotificationsConfigError(Error):
	'''
	Exception raised when we try to access the mails notifications configuration without any configuration file provided.
	'''

	pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import codecs
import base64
import hashlib
import json

def intOrNone(s):
	'''
	Convert a string into either an integer or a `None`.

	Parameters
	----------
	s : str
		The string to convert.

	Returns
	-------
	converted : int|None
		The wanted integer if `s` represents one, `None` otherwise.
	'''

	if re.match(r'^-?[0-9]+$', s):
		return int(s)

	return None

def fromObject(obj):
	'''
	Represent an object as a base64 string.

	Parameters
	----------
	obj : dict|list
		The object to represent.

	Returns
	-------
	obj_str : str
		The corresponding base64 string.
	'''

	obj_str = json.dumps(obj, sort_keys = True)
	obj_str = base64.urlsafe_b64encode(obj_str.encode('utf-8')).decode('ascii')
	return obj_str

def hash(s):
	'''
	Hash a string, mostly to serve as identifier.

	Parameters
	----------
	s : str
		The string to hash.

	Returns
	-------
	hash : str
		The hash.
	'''

	h = hashlib.md5(s.encode('utf-8')).hexdigest()
	b = codecs.encode(codecs.decode(h.encode('utf-8'), 'hex'), 'base64').decode()

	return b[:-1].replace('+', '-').replace('/', '_')[:-2]

def plural(n, if_single, if_plural, *, add_n = True):
	'''
	Return a string or another, depending on a number.

	Parameters
	----------
	n : int
		The number to test.

	if_single : str
		The string to return if `n` is lower or equal than 1.

	if_plural : str
		The string to return if `n` is greater than 1.

	add_n : bool
		`True` to prepend the string by the number, `False` to only get the string.

	Returns
	-------
	s : str
		The single or plural string, depending on the test.
	'''

	s = if_plural if n > 1 else if_single

	if add_n:
		s = f'{n} {s}'

	return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

from . import jsonfiles

class Folder():
	'''
	Base class for each system needing access to the configuration files of a simulations folder.
	Initialize with the simulations folder and load the settings.

	Parameters
	----------
	folder : str
		The simulations folder. Must contain a settings file.

	Raises
	------
	FileNotFoundError
		No `.simulations.conf` file found in folder.
	'''

	def __init__(self, folder):
		self._folder = folder
		self._settings_file = os.path.join(self._folder, '.simulations.conf')

		if not(os.path.isfile(self._settings_file)):
			raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self._settings_file)

		self._settings = None

	@property
	def folder(self):
		'''
		Return the folder's path.

		Returns
		-------
		path : str
			The path.
		'''

		return self._folder

	@property
	def settings(self):
		'''
		Return the content of the settings file as a dictionary.

		Returns
		-------
		settings : dict
			The folder's settings.
		'''

		if not(self._settings):
			self._settings = jsonfiles.read(self._settings_file)

		return self._settings



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import stat
import time
import tempfile

from . import string

from .folder import Folder
from .simulation import Simulation
from .manager import Manager
from .generator import Generator
from .remote import RemoteFolder
from .watcher import Watcher
from .ui import UI
from .errors import *

class Maker():
	'''
	Assemble all components to extract simulations and automatically create them if they don't exist.

	Parameters
	----------
	simulations_folder : str
		The simulations folder. Must contain a settings file.

	remote_folder_conf : dict
		Configuration of the remote folder.

	mail_config : dict
		Configuration of the mailbox.

	mail_notifications_config : dict
		Mails notifications configuration.

	max_corrupted : int
		Maximum number of allowed corruptions. Corruptions counter is incremented each time at least one simulation is corrupted. If negative, there is no limit.

	max_failures : int
		Maximum number of allowed failures in the execution of a job. The counter is incremented each time at least one job fails. If negative, there is no limit.

	ui : bool
		`True` to show different informations in the UI, `False` to show nothing.
	'''

	def __init__(self, simulations_folder, remote_folder_conf, *, mail_config = None, mail_notifications_config = None, max_corrupted = -1, max_failures = 0, ui = False):
		self._simulations_folder = Folder(simulations_folder)
		self._remote_folder_conf = remote_folder_conf

		self._manager_instance = None
		self._generator_instance = None
		self._remote_folder_instance = None
		self._watcher_instance = None
		self._ui_instance = None
		self._ui_state_line = None

		self._mail_config = mail_config
		self._mail_notifications_config = mail_notifications_config

		self._max_corrupted = max_corrupted
		self._corruptions_counter = 0

		self._max_failures = max_failures
		self._failures_counter = 0

		self._show_ui = ui

	@property
	def _manager(self):
		'''
		Returns the instance of Manager used in the Maker.

		Returns
		-------
		manager : Manager
			Current instance, or a new one if `None`.
		'''

		if not(self._manager_instance):
			self._manager_instance = Manager(self._simulations_folder)

		return self._manager_instance

	@property
	def _generator(self):
		'''
		Returns the instance of Generator used in the Maker.

		Returns
		-------
		generator : Generator
			Current instance, or a new one if `None`.
		'''

		if not(self._generator_instance):
			self._generator_instance = Generator(self._simulations_folder)

		return self._generator_instance

	@property
	def _remote_folder(self):
		'''
		Returns the instance of RemoteFolder used in the Maker.

		Returns
		-------
		remote_folder : RemoteFolder
			Current instance, or a new one if `None`.
		'''

		if not(self._remote_folder_instance):
			self._remote_folder_instance = RemoteFolder(self._remote_folder_conf)
			self.displayState('Connecting to the folder…')
			self._remote_folder_instance.open()
			self.displayState('Connection done.')

		return self._remote_folder_instance

	@property
	def _watcher(self):
		'''
		Returns the instance of Watcher used in the Maker.

		Returns
		-------
		watcher : Watcher
			Current instance, or a new one if `None`.
		'''

		if not(self._watcher_instance):
			if self._mail_config is None:
				self._watcher_instance = Watcher(remote_folder = self._remote_folder)

			else:
				self._watcher_instance = Watcher(mail_config = self._mail_config, mail_notifications_config = self._mail_notifications_config)

		return self._watcher_instance

	@property
	def _ui(self):
		'''
		Returns the instance of UI used in the Maker.

		Returns
		-------
		ui : UI
			Current instance, or a new one if `None`.
		'''

		if not(self._ui_instance) and self._show_ui:
			self._ui_instance = UI()

		return self._ui_instance

	def close(self):
		'''
		Clear all instances of the modules.
		'''

		self._manager_instance = None
		self._generator_instance = None

		try:
			self._remote_folder_instance.close()

		except AttributeError:
			pass

		self._remote_folder_instance = None

		try:
			self._watcher_instance.close()

		except AttributeError:
			pass

		self._watcher_instance = None

	def displayTextLine(self, text):
		'''
		Display a new text line.

		Parameters
		----------
		text : str
			Text to display.

		Returns
		-------
		text_line : str
			ID of the text line.
		'''

		if not(self._show_ui):
			return None

		return self._ui.addTextLine(text)

	def updateTextLine(self, text_line, new_text):
		'''
		Update a text line.

		Parameters
		----------
		text_line : str
			ID of the text line to update.

		new_text : str
			New text to display.
		'''

		if not(self._show_ui):
			return

		self._ui.replaceTextLine(text_line, new_text)

	def removeTextLine(self, text_line):
		'''
		Remove a text line.

		Parameters
		----------
		text_line : str
			ID of the text line to remove.
		'''

		if not(self._show_ui):
			return

		self._ui.removeTextLine(text_line)

	def displayState(self, state):
		'''
		Display a state message (creation of an instance, connection, …).

		Parameters
		----------
		state : str
			State to display.
		'''

		if not(self._show_ui):
			return

		if self._ui_state_line is None:
			self._ui_state_line = self.displayTextLine(state)

		else:
			self.updateTextLine(self._ui_state_line, state)

	def displayProgressBar(self, N):
		'''
		Add a new progress bar.

		Parameters
		----------
		N : int
			Final number to reach.

		Returns
		-------
		progress_bar : str
			ID of the progress bar.
		'''

		if not(self._show_ui):
			return None

		return self._ui.addProgressBar(N)

	def updateProgressBar(self, progress_bar, n = None):
		'''
		Update a progress bar.

		Parameters
		----------
		progress_bar : str
			ID of the progress bar to update.

		n : int
			New number to display (`None` to increment by one).
		'''

		if not(self._show_ui):
			return

		self._ui.updateProgressBar(progress_bar, n)

	def removeProgressBar(self, progress_bar):
		'''
		Remove a progress bar.

		Parameters
		----------
		progress_bar : str
			ID of the progress bar to remove.
		'''

		if not(self._show_ui):
			return

		self._ui.removeProgressBar(progress_bar)

	def parseScriptToLaunch(self, launch_option):
		'''
		Parse the `launch` option of a recipe to determine which skeleton/script must be called.

		Parameters
		----------
		launch_option : str
			Value of the `launch` option to parse.

		Returns
		-------
		script_to_launch : dict
			"Coordinates" of the script to launch.
		'''

		self.displayState('Parsing the launch option…')

		option_split = launch_option.rsplit(':', maxsplit = 2)
		option_split_num = [string.intOrNone(s) for s in option_split]

		cut = max([k for k, n in enumerate(option_split_num) if n is None]) + 1

		coords = option_split_num[cut:]
		coords += [-1] * (2 - len(coords))

		return {
			'name': ':'.join(option_split[:cut]),
			'skeleton': coords[0],
			'script': coords[1]
		}

	def run(self, simulations, generator_recipe):
		'''
		Main loop, run until all simulations are extracted or some jobs failed.

		Parameters
		----------
		simulations : list
			List of simulations to extract/generate.

		generator_recipe : dict
			Recipe to use in the generator to generate the scripts.

		Returns
		-------
		unknown_simulations : list
			List of simulations that failed to be generated.
		'''

		script_coords = self.parseScriptToLaunch(generator_recipe['launch'])

		self._corruptions_counter = 0
		self._failures_counter = 0

		while (self._max_corrupted < 0 or self._corruptions_counter <= self._max_corrupted) and (self._max_failures < 0 or self._failures_counter <= self._max_failures):
			unknown_simulations = self.extractSimulations(simulations)

			if not(unknown_simulations):
				break

			jobs_ids = self.generateSimulations(unknown_simulations, generator_recipe, script_coords)
			success = self.waitForJobs(jobs_ids, generator_recipe)

			if not(success):
				self._failures_counter += 1

			success = self.downloadSimulations(unknown_simulations)

			if not(success):
				self._corruptions_counter += 1

			self.displayState('Deleting the scripts folder…')
			self._remote_folder.deleteRemote([generator_recipe['basedir']])

		if unknown_simulations:
			self.displayState(string.plural(len(unknown_simulations), 'simulation still does not exist', 'simulations still do not exist'))

		else:
			self.displayState('All simulations have successfully been extracted')

		return unknown_simulations

	def extractSimulations(self, simulations):
		'''
		Extract a given set of simulations.

		Parameters
		----------
		simulations : list
			The list of simulations to extract.

		Returns
		-------
		unknown_simulations : list
			The list of simulations which do not exist (yet).
		'''

		self.displayState('Extracting the simulations…')
		progress_bar = self.displayProgressBar(len(simulations))

		unknown_simulations = self._manager.batchExtract(simulations, callback = lambda : self.updateProgressBar(progress_bar))

		self.removeProgressBar(progress_bar)

		return unknown_simulations

	def generateSimulations(self, simulations, recipe, script_coords):
		'''
		Generate the scripts to generate some unknown simulations, and run them.

		Parameters
		----------
		simulations : list
			List of simulations to create.

		recipe : dict
			Recipe to use in the generator.

		script_coords : dict
			'Coordinates' of the script to launch.

		Raises
		------
		ScriptNotFoundError
			The script to launch has not been found.

		Returns
		-------
		jobs_ids : list
			IDs of the jobs to wait.
		'''

		self.displayState('Generating the scripts…')

		scripts_dir = tempfile.mkdtemp(prefix = 'simulations-scripts_')
		recipe['basedir'] = self._remote_folder.sendDir(scripts_dir)

		self._generator.add(simulations)
		generated_scripts = self._generator.generate(scripts_dir, recipe, empty_dest = True)
		self._generator.clear()

		possible_skeletons_to_launch = [k for k, s in enumerate(recipe['subgroups_skeletons'] + recipe['wholegroup_skeletons']) if s == script_coords['name']]

		try:
			script_to_launch = generated_scripts[possible_skeletons_to_launch[script_coords['skeleton']]][script_coords['script']]

		except IndexError:
			raise ScriptNotFoundError(script_coords)

		script_mode = os.stat(script_to_launch['localpath']).st_mode
		if not(script_mode & stat.S_IXUSR & stat.S_IXGRP & stat.S_IXOTH):
			os.chmod(script_to_launch['localpath'], script_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

		self._remote_folder.sendDir(scripts_dir, delete = True, empty_dest = True)

		output = self._remote_folder.execute(script_to_launch['finalpath'])
		jobs_ids = list(map(lambda l: l.strip(), output.readlines()))

		return jobs_ids

	def waitForJobs(self, jobs_ids, recipe):
		'''
		Wait for a given list of jobs to finish.

		Parameters
		----------
		jobs_ids : list
			IDs of the jobs to wait.

		recipe : dict
			Generator recipe to use.

		Returns
		-------
		success : bool
			`True` is all jobs were finished normally, `False` if there was at least one failure.
		'''

		jobs_number = len(jobs_ids)
		jobs_numbers_by_state = {}

		self.displayState('Waiting for jobs to finish…')
		progress_bar = self.displayProgressBar(jobs_number)

		statuses = 'Current statuses: {waiting} waiting, {running} running, {succeed} succeed, {failed} failed'
		statuses_line = self.displayTextLine('')

		self._watcher.addJobsToWatch(jobs_ids)

		if self._mail_config is None:
			self._watcher.setJobsStatesPath(recipe['jobs_states_filename'])

		while True:
			self._watcher.updateJobsStates()
			jobs_numbers_by_state = self._watcher.getNumberOfJobsByStates(['waiting', 'running', 'succeed', 'failed'])
			finished = jobs_numbers_by_state['succeed'] + jobs_numbers_by_state['failed']

			self.updateProgressBar(progress_bar, finished)
			self.updateTextLine(statuses_line, statuses.format(**jobs_numbers_by_state))

			if finished == jobs_number:
				break

			time.sleep(0.5)

		self.removeProgressBar(progress_bar)
		self.removeTextLine(statuses_line)

		self._watcher.clearJobs()

		return jobs_numbers_by_state['failed'] == 0

	def downloadSimulations(self, simulations):
		'''
		Download the generated simulations and add them to the manager.

		Parameters
		----------
		simulations : list
			List of simulations to download.

		Returns
		-------
		success : bool
			`True` if all simulations has successfully been downloaded and added, `False` if there has been at least one issue.
		'''

		self.displayState('Downloading the simulations…')
		progress_bar = self.displayProgressBar(len(simulations))

		simulations_to_add = []

		for simulation in simulations:
			simulation = Simulation.ensureType(simulation, self._simulations_folder)

			tmpdir = tempfile.mkdtemp(prefix = 'simulation_')
			try:
				self._remote_folder.receiveDir(simulation['folder'], tmpdir, delete = True)

			except RemotePathNotFoundError:
				pass

			simulation['folder'] = tmpdir
			simulations_to_add.append(simulation)

			self.updateProgressBar(progress_bar)

		self.removeProgressBar(progress_bar)

		self.displayState('Adding the simulations to the manager…')
		progress_bar = self.displayProgressBar(len(simulations))

		failed_to_add = self._manager.batchAdd(simulations_to_add, callback = lambda : self.updateProgressBar(progress_bar))

		self.removeProgressBar(progress_bar)

		return not(bool(failed_to_add))
